/* AM modulated signal generator
 */

#include <avr/interrupt.h>
#include <avr/io.h>
#include <setjmp.h>
#include <string.h>
#include <avr/pgmspace.h>
#include "sig455.h"
#include "uart.h"
#include "timers.h"

// Square wave with 50% duty-cycle
unsigned char sw50[256] __attribute__ ((aligned(256)));
// Square wave with 10% duty-cycle
unsigned char sw10[256] __attribute__ ((aligned(256)));
jmp_buf jmpb;
unsigned long freq;
unsigned long dphase1 = 0x000e6666; // 100K @ 16MHZ
unsigned long dphase2 = 0x000e6666; // 100K @ 16MHZ
uint8_t tmod = 77; // 400Hz
uint8_t tbit = 0;;

#ifdef	WITHOUT_DAC
/* RF on PB5, Arduino: D13
 * LF on PD6, Arduino: D6 (OC0A)
 */
void
gen_tables()
{
	unsigned char *dst;
	int n;

	dst = sw50;
	for (n = 0 ; n < 128 ; n++)
		*dst++ = 255;
	for (n = 128 ; n < 256 ; n++)
		*dst++ = 0;

	dst = sw10;
	for (n = 0 ; n < 32 ; n++)
		*dst++ = 255;
	for (n = 32 ; n < 256 ; n++)
		*dst++ = 0;
}
#endif

#ifdef	WITH_DAC
void
gen_tables()
{
	unsigned char *dst;
	int n;

	dst = sw50;
	for (n = 0 ; n < 128 ; n++)
		*dst++ = 255;
	for (n = 128 ; n < 256 ; n++)
		*dst++ = 0;

	dst = sw10;
	for (n = 0 ; n < 32 ; n++)
		*dst++ = 146;
	for (n = 32 ; n < 256 ; n++)
		*dst++ = 110;
}
#endif

int
main()
{
	char cmd,ch;
	unsigned long param;
	char mode = 'a';

	/* Port directions
	 * PORTB: 7-0 output
	 * PORTD: 7-1 output, 0 input
	 */
	DDRB = 0xff;
	DDRD = 0xfe;

	uart_init();
	timer0_init();

	gen_tables();

	// return here after keyboard interrupt
	if (setjmp(jmpb) != 0) {
		cli();
	} else {
		sei();
		gen_am(dphase1);
	}

	uart_puts("\r\nAM/FM signal generator\r\n");

	/* Read command */
	uart_putc('>');
	cmd = uart_getc();
	uart_putc(cmd);
	param = 0L;
	while ((ch = uart_getc()) != '\r') {
		param = param * 10L + (long)ch - (long)'0';
		uart_putc(ch);
	}
	switch (cmd) {
	case 'a':	// Frequency in AM mode
		freq = param;
		mode = 'a';
		// 24 bit accumulator
		// dds loop : 9 cycles
		dphase1 = (long long) freq * 0x00ffffffUL * 9UL / F_CPU;
		break;
	case 'f':	// Frequency in FM mode
		freq = param;
		mode = 'f';
		// 24 bit accumulator
		// dds loop : 9 cycles
		dphase1 = (long long) freq * 0x00ffffffUL * 9UL / F_CPU;
		dphase2 = (long long) (freq + 5000UL) * 0x00ffffffUL * 9UL / F_CPU;
		// Enable Timer Overflow Interrupt
		TIMSK0 |= (1 << OCIE0A);
		break;
	case 'm':	// Modulation
		if (param == 0) {
			// Disable timer interrupt
			TIMSK0 &= ~(1 << OCIE0A);
		} else {
			// OCR0A = (F_CPU / (2 * prescale * Fout)) - 1
			// prescale: 256
			tmod = F_CPU / ( 2 * 256 * param) - 1;
			TIMSK0 |= (1 << OCIE0A);
		}
		// Enable Timer Overflow Interrupt
		TIMSK0 |= (1 << OCIE0A);
		break;
	case 't':	// Toggle T bit, disable timer interrupt
		TIMSK0 &= ~(1 << OCIE0A);
		tbit ^= 0x40;
		break;
	}
	uart_puts("\r\nRunning .... ");
	uart_hexlong(dphase1);
	uart_putc(' ');
	uart_hexlong(dphase2);
	uart_putc(' ');
	uart_hexbyte(tmod);

	// set T bit just before entering dds loop
	// T bit is used by __umul... and __div routines
	OCR0A = tmod;
	if (tbit) {
		SREG |= 0x40;
	} else {
		SREG &= ~0x40;
	}
	if (mode == 'a') {
		sei();
		gen_am(dphase1);
	} else {
		sei();
		gen_fm(dphase1, dphase2);
	}
}

